@@ -1,6 +1,6 @@
 /* libhttpd.h - defines for libhttpd
 **
-** Copyright  1995,1998,1999,2000 by Jef Poskanzer <jef@acme.com>.
+** Copyright  1995,1998,1999,2000,2001 by Jef Poskanzer <jef@acme.com>.
 ** All rights reserved.
 **
 ** Redistribution and use in source and binary forms, with or without
@@ -72,9 +72,11 @@ typedef struct {
     char* charset;
     char* cwd;
     int listen4_fd, listen6_fd;
+    int no_log;
     FILE* logfp;
     int no_symlink;
     int vhost;
+    int global_passwd;
     char* url_pattern;
     char* local_pattern;
     int no_empty_referers;
@@ -90,7 +92,8 @@ typedef struct {
     int checked_state;
     int method;
     int status;
-    off_t bytes;
+    off_t bytes_to_send;
+    off_t bytes_sent;
     char* encodedurl;
     char* decodedurl;
     char* protocol;
@@ -103,6 +106,7 @@ typedef struct {
     char* useragent;
     char* accept;
     char* accepte;
+    char* acceptl;
     char* cookie;
     char* contenttype;
     char* reqhost;
@@ -136,6 +140,7 @@ typedef struct {
     } httpd_conn;
 
 /* Methods. */
+#define METHOD_UNKNOWN 0
 #define METHOD_GET 1
 #define METHOD_HEAD 2
 #define METHOD_POST 3
@@ -146,12 +151,13 @@ typedef struct {
 #define CHST_SECONDWORD 2
 #define CHST_SECONDWS 3
 #define CHST_THIRDWORD 4
-#define CHST_LINE 5
-#define CHST_LF 6
-#define CHST_CR 7
-#define CHST_CRLF 8
-#define CHST_CRLFCR 9
-#define CHST_BOGUS 10
+#define CHST_THIRDWS 5
+#define CHST_LINE 6
+#define CHST_LF 7
+#define CHST_CR 8
+#define CHST_CRLF 9
+#define CHST_CRLFCR 10
+#define CHST_BOGUS 11
 
 
 /* Initializes.  Does the socket(), bind(), and listen().   Returns an
@@ -160,8 +166,9 @@ typedef struct {
 */
 extern httpd_server* httpd_initialize(
     char* hostname, httpd_sockaddr* sa4P, httpd_sockaddr* sa6P, int port,
-    char* cgi_pattern, char* charset, char* cwd, FILE* logfp, int no_symlink,
-    int vhost, char* url_pattern, char* local_pattern, int no_empty_referers );
+    char* cgi_pattern, char* charset, char* cwd, int no_log, FILE* logfp,
+    int no_symlink, int vhost, int global_passwd, char* url_pattern,
+    char* local_pattern, int no_empty_referers );
 
 /* Change the log file. */
 extern void httpd_set_logfp( httpd_server* hs, FILE* logfp );
@@ -206,11 +213,12 @@ extern int httpd_parse_request( httpd_conn* hc );
 /* Starts sending data back to the client.  In some cases (directories,
 ** CGI programs), finishes sending by itself - in those cases, hc->file_fd
 ** is <0.  If there is more data to be sent, then hc->file_fd is a file
-** descriptor for the file to send.
+** descriptor for the file to send.  If you don't have a current timeval
+** handy just pass in 0.
 **
 ** Returns -1 on error.
 */
-extern int httpd_start_request( httpd_conn* hc );
+extern int httpd_start_request( httpd_conn* hc, struct timeval* nowP );
 
 /* Actually sends any buffered response text. */
 extern void httpd_write_response( httpd_conn* hc );
@@ -240,13 +248,22 @@ extern char* httpd_err408form;
 extern char* httpd_err503title;
 extern char* httpd_err503form;
 
-/* Generates a string representation of a method number. */
+/* Generate a string representation of a method number. */
 extern char* httpd_method_str( int method );
 
-/* Reallocates a string. */
+/* Reallocate a string. */
 extern void httpd_realloc_str( char** strP, int* maxsizeP, int size );
 
-/* Formats a network socket to a string representation. */
+/* Format a network socket to a string representation. */
 extern char* httpd_ntoa( httpd_sockaddr* saP );
 
+/* Set NDELAY mode on a socket. */
+void httpd_set_ndelay( int fd );
+
+/* Clear NDELAY mode on a socket. */
+void httpd_clear_ndelay( int fd );
+
+/* Generate debugging statistics syslog message. */
+extern void httpd_logstats( long secs );
+
 #endif /* _LIBHTTPD_H_ */
