@@ -1,6 +1,6 @@
 /* mmc.c - mmap cache
 **
-** Copyright  1998 by Jef Poskanzer <jef@acme.com>.
+** Copyright  1998,2001 by Jef Poskanzer <jef@acme.com>.
 ** All rights reserved.
 **
 ** Redistribution and use in source and binary forms, with or without
@@ -40,16 +40,34 @@
 #include "mmc.h"
 
 
+/* Defines. */
+#ifndef DEFAULT_EXPIRE_AGE
+#define DEFAULT_EXPIRE_AGE 600
+#endif
+#ifndef DESIRED_FREE_COUNT
+#define DESIRED_FREE_COUNT 100
+#endif
+#ifndef DESIRED_MAX_MAPPED_FILES
+#define DESIRED_MAX_MAPPED_FILES 2000
+#endif
+#ifndef INITIAL_HASH_SIZE
+#define INITIAL_HASH_SIZE (1 << 10)
+#endif
+
+#define max(a,b) ((a)>(b)?(a):(b))
+#define min(a,b) ((a)<(b)?(a):(b))
+
+
 /* The Map struct. */
 typedef struct MapStruct {
     ino_t ino;
     dev_t dev;
     off_t size;
-    time_t mtime;
+    time_t ctime;
     int refcount;
     time_t reftime;
     void* addr;
-    int hash;
+    unsigned int hash;
     int hash_idx;
     struct MapStruct* next;
     } Map;
@@ -57,42 +75,32 @@ typedef struct MapStruct {
 
 /* Globals. */
 static Map* maps = (Map*) 0;
-static int map_count = 0;
 static Map* free_maps = (Map*) 0;
-static int free_count = 0;
+static int alloc_count = 0, map_count = 0, free_count = 0;
 static Map** hash_table = (Map**) 0;
 static int hash_size;
+static unsigned int hash_mask;
+static time_t expire_age = DEFAULT_EXPIRE_AGE;
 
 
-/* Defines. */
-#ifndef EXPIRE_AGE
-#define EXPIRE_AGE 600
-#endif
-#ifndef DESIRED_FREE_COUNT
-#define DESIRED_FREE_COUNT 100
-#endif
-#ifndef INITIAL_HASH_SIZE
-#define INITIAL_HASH_SIZE 1009
-#endif
-
 
 /* Forwards. */
 static void really_unmap( Map** mm );
 static int check_hash_size( void );
-static int is_prime( int n );
 static int add_hash( Map* m );
-static Map* find_hash( ino_t ino, dev_t dev, off_t size, time_t mtime );
-static int hash( ino_t ino, dev_t dev, off_t size, time_t mtime );
+static Map* find_hash( ino_t ino, dev_t dev, off_t size, time_t ctime );
+static unsigned int hash( ino_t ino, dev_t dev, off_t size, time_t ctime );
 
 
 void*
-mmc_map( char* filename, struct stat* sbP )
+mmc_map( char* filename, struct stat* sbP, struct timeval* nowP )
     {
+    time_t now;
     struct stat sb;
     Map* m;
     int fd;
 
-    /* Stat the file if necessary. */
+    /* Stat the file, if necessary. */
     if ( sbP != (struct stat*) 0 )
 	sb = *sbP;
     else
@@ -104,21 +112,28 @@ mmc_map( char* filename, struct stat* sbP )
 	    }
 	}
 
+    /* Get the current time, if necessary. */
+    if ( nowP != (struct timeval*) 0 )
+	now = nowP->tv_sec;
+    else
+	now = time( (time_t*) 0 );
+
     /* See if we have it mapped already, via the hash table. */
     if ( check_hash_size() < 0 )
 	{
 	syslog( LOG_ERR, "check_hash_size() failure" );
 	return (void*) 0;
 	}
-    m = find_hash( sb.st_ino, sb.st_dev, sb.st_size, sb.st_mtime );
+    m = find_hash( sb.st_ino, sb.st_dev, sb.st_size, sb.st_ctime );
     if ( m != (Map*) 0 )
 	{
-	/* Yep. */
+	/* Yep.  Just return the existing map */
 	++m->refcount;
+	m->reftime = now;
 	return m->addr;
 	}
 
-    /* Nope.  Open the file. */
+    /* Open the file. */
     fd = open( filename, O_RDONLY );
     if ( fd < 0 )
 	{
@@ -139,16 +154,19 @@ mmc_map( char* filename, struct stat* sbP )
 	if ( m == (Map*) 0 )
 	    {
 	    (void) close( fd );
+	    syslog( LOG_ERR, "out of memory allocating a Map" );
 	    return (void*) 0;
 	    }
+	++alloc_count;
 	}
 
     /* Fill in the Map entry. */
     m->ino = sb.st_ino;
     m->dev = sb.st_dev;
     m->size = sb.st_size;
-    m->mtime = sb.st_mtime;
+    m->ctime = sb.st_ctime;
     m->refcount = 1;
+    m->reftime = now;
 
     /* Avoid doing anything for zero-length files; some systems don't like
     ** to mmap them, other systems dislike mallocing zero bytes.
@@ -165,6 +183,7 @@ mmc_map( char* filename, struct stat* sbP )
 	    syslog( LOG_ERR, "mmap - %m" );
 	    (void) close( fd );
 	    free( (void*) m );
+	    --alloc_count;
 	    return (void*) 0;
 	    }
 #else /* HAVE_MMAP */
@@ -172,9 +191,10 @@ mmc_map( char* filename, struct stat* sbP )
 	m->addr = (void*) malloc( m->size );
 	if ( m->addr == (void*) 0 )
 	    {
-	    syslog( LOG_ERR, "not enough memory" );
+	    syslog( LOG_ERR, "out of memory storing a file" );
 	    (void) close( fd );
 	    free( (void*) m );
+	    --alloc_count;
 	    return (void*) 0;
 	    }
 	if ( read( fd, m->addr, m->size ) != m->size )
@@ -182,6 +202,7 @@ mmc_map( char* filename, struct stat* sbP )
 	    syslog( LOG_ERR, "read - %m" );
 	    (void) close( fd );
 	    free( (void*) m );
+	    --alloc_count;
 	    return (void*) 0;
 	    }
 #endif /* HAVE_MMAP */
@@ -193,6 +214,7 @@ mmc_map( char* filename, struct stat* sbP )
 	{
 	syslog( LOG_ERR, "add_hash() failure" );
 	free( (void*) m );
+	--alloc_count;
 	return (void*) 0;
 	}
 
@@ -207,25 +229,33 @@ mmc_map( char* filename, struct stat* sbP )
 
 
 void
-mmc_unmap( void* addr, struct timeval* nowP, struct stat* sbP )
+mmc_unmap( void* addr, struct stat* sbP, struct timeval* nowP )
     {
-    Map* m;
+    Map* m = (Map*) 0;
 
-    /* Find the Map entry for this address, either by hashing or searching. */
+    /* Find the Map entry for this address.  First try a hash. */
     if ( sbP != (struct stat*) 0 )
-	m = find_hash( sbP->st_ino, sbP->st_dev, sbP->st_size, sbP->st_mtime );
-    else
+	{
+	m = find_hash( sbP->st_ino, sbP->st_dev, sbP->st_size, sbP->st_ctime );
+	if ( m != (Map*) 0 && m->addr != addr )
+	    m = (Map*) 0;
+	}
+    /* If that didn't work, try a full search. */
+    if ( m == (Map*) 0 )
 	for ( m = maps; m != (Map*) 0; m = m->next )
 	    if ( m->addr == addr )
 		break;
-    if ( m != (Map*) 0 )
+    if ( m == (Map*) 0 )
+	syslog( LOG_ERR, "mmc_unmap failed to find entry!" );
+    else if ( m->refcount <= 0 )
+	syslog( LOG_ERR, "mmc_unmap found zero or negative refcount!" );
+    else
 	{
 	--m->refcount;
 	if ( nowP != (struct timeval*) 0 )
 	    m->reftime = nowP->tv_sec;
 	else
 	    m->reftime = time( (time_t*) 0 );
-	return;
 	}
     }
 
@@ -237,22 +267,28 @@ mmc_cleanup( struct timeval* nowP )
     Map** mm;
     Map* m;
 
-    /* Get current time, if necessary. */
+    /* Get the current time, if necessary. */
     if ( nowP != (struct timeval*) 0 )
 	now = nowP->tv_sec;
     else
 	now = time( (time_t*) 0 );
 
-    /* Really unmap any unreferenced entries older than the limit. */
+    /* Really unmap any unreferenced entries older than the age limit. */
     for ( mm = &maps; *mm != (Map*) 0; )
 	{
 	m = *mm;
-	if ( m->refcount == 0 && now - m->reftime >= EXPIRE_AGE )
+	if ( m->refcount == 0 && now - m->reftime >= expire_age )
 	    really_unmap( mm );
 	else
 	    mm = &(*mm)->next;
 	}
 
+    /* If there are still too many (or too few) maps, adjust the age limit. */
+    if ( map_count > DESIRED_MAX_MAPPED_FILES )
+	expire_age = max( ( expire_age * 2 ) / 3, DEFAULT_EXPIRE_AGE / 10 );
+    else if ( map_count < DESIRED_MAX_MAPPED_FILES / 2 )
+	expire_age = min( ( expire_age * 5 ) / 4, DEFAULT_EXPIRE_AGE * 3 );
+
     /* Really free excess blocks on the free list. */
     while ( free_count > DESIRED_FREE_COUNT )
 	{
@@ -260,6 +296,7 @@ mmc_cleanup( struct timeval* nowP )
 	free_maps = m->next;
 	--free_count;
 	free( (void*) m );
+	--alloc_count;
 	}
     }
 
@@ -285,8 +322,8 @@ really_unmap( Map** mm )
     m->next = free_maps;
     free_maps = m;
     ++free_count;
-    /* This will sometimes break hash chains, but that's harmless.
-    ** Worst case we map another copy of the same file.
+    /* This will sometimes break hash chains, but that's harmless; the
+    ** unmapping code that searches the hash table knows to keep searching.
     */
     hash_table[m->hash_idx] = (Map*) 0;
     }
@@ -305,18 +342,11 @@ mmc_destroy( void )
 	free_maps = m->next;
 	--free_count;
 	free( (void*) m );
+	--alloc_count;
 	}
     }
 
 
-void
-mmc_stats( int* activeP, int* freeP )
-    {
-    *activeP = map_count;
-    *freeP = free_count;
-    }
-
-
 /* Make sure the hash table is big enough. */
 static int
 check_hash_size( void )
@@ -326,7 +356,10 @@ check_hash_size( void )
 
     /* Are we just starting out? */
     if ( hash_table == (Map**) 0 )
+	{
 	hash_size = INITIAL_HASH_SIZE;
+	hash_mask = hash_size - 1;
+	}
     /* Is it at least three times bigger than the number of entries? */
     else if ( hash_size >= map_count * 3 )
 	return 0;
@@ -334,9 +367,13 @@ check_hash_size( void )
 	{
 	/* No, got to expand. */
 	free( (void*) hash_table );
-	/* Find a prime at least six times bigger. */
-	for ( hash_size = map_count * 6; ! is_prime( hash_size ); ++hash_size )
-	    continue;
+	/* Double the hash size until it's big enough. */
+	do
+	    {
+	    hash_size = hash_size << 1;
+	    }
+	while ( hash_size < map_count * 6 );
+	hash_mask = hash_size - 1;
 	}
     /* Make the new table. */
     hash_table = (Map**) malloc( hash_size * sizeof(Map*) );
@@ -353,32 +390,14 @@ check_hash_size( void )
     }
 
 
-static int
-is_prime( int n )
-    {
-    int m, nom;
-
-    if ( n % 2 == 0 )
-	return 0;
-    for ( m = 3; ; m += 2 )
-	{
-	nom = n / m;
-	if ( nom < m )
-	    return 1;
-	if ( nom * m == n )
-	    return 0;
-	}
-    }
-
-
 static int
 add_hash( Map* m )
     {
-    int h, he, i;
+    unsigned int h, he, i;
 
-    h = hash( m->ino, m->dev, m->size, m->mtime );
-    he = ( h + hash_size - 1 ) % hash_size;
-    for ( i = h; ; i = ( i + 1 ) % hash_size )
+    h = hash( m->ino, m->dev, m->size, m->ctime );
+    he = ( h + hash_size - 1 ) & hash_mask;
+    for ( i = h; ; i = ( i + 1 ) & hash_mask )
 	{
 	if ( hash_table[i] == (Map*) 0 )
 	    {
@@ -395,20 +414,20 @@ add_hash( Map* m )
 
 
 static Map*
-find_hash( ino_t ino, dev_t dev, off_t size, time_t mtime )
+find_hash( ino_t ino, dev_t dev, off_t size, time_t ctime )
     {
-    int h, he, i;
+    unsigned int h, he, i;
     Map* m;
 
-    h = hash( ino, dev, size, mtime );
-    he = ( h + hash_size - 1 ) % hash_size;
-    for ( i = h; ; i = ( i + 1 ) % hash_size )
+    h = hash( ino, dev, size, ctime );
+    he = ( h + hash_size - 1 ) & hash_mask;
+    for ( i = h; ; i = ( i + 1 ) & hash_mask )
 	{
 	m = hash_table[i];
 	if ( m == (Map*) 0 )
 	    break;
 	if ( m->hash == h && m->ino == ino && m->dev == dev &&
-	     m->size == size && m->mtime == mtime )
+	     m->size == size && m->ctime == ctime )
 	    return m;
 	if ( i == he )
 	    break;
@@ -417,8 +436,30 @@ find_hash( ino_t ino, dev_t dev, off_t size, time_t mtime )
     }
 
 
-static int
-hash( ino_t ino, dev_t dev, off_t size, time_t mtime )
+static unsigned int
+hash( ino_t ino, dev_t dev, off_t size, time_t ctime )
     {
-    return ( ino ^ dev ^ size ^ mtime ) % hash_size;
+    unsigned int h = 177573;
+
+    h ^= ino;
+    h += h << 5;
+    h ^= dev;
+    h += h << 5;
+    h ^= size;
+    h += h << 5;
+    h ^= ctime;
+
+    return h & hash_mask;
+    }
+
+
+/* Generate debugging statistics syslog message. */
+void
+mmc_logstats( long secs )
+    {
+    syslog(
+	LOG_NOTICE, "  map cache - %d allocated, %d active, %d free; hash size: %d; expire age: %ld",
+	alloc_count, map_count, free_count, hash_size, expire_age );
+    if ( map_count + free_count != alloc_count )
+	syslog( LOG_ERR, "map counts don't add up!" );
     }
