@@ -30,14 +30,15 @@
 
 /* Returns an mmap()ed area for the given file, or (void*) 0 on errors.
 ** If you have a stat buffer on the file, pass it in, otherwise pass 0.
+** Same for the current time.
 */
-extern void* mmc_map( char* filename, struct stat* sbP );
+extern void* mmc_map( char* filename, struct stat* sbP, struct timeval* nowP );
 
 /* Done with an mmap()ed area that was returned by mmc_map().
-** If you have the current time, pass it in, otherwise pass 0.
-** Same for the stat buffer.
+** If you have a stat buffer on the file, pass it in, otherwise pass 0.
+** Same for the current time.
 */
-extern void mmc_unmap( void* addr, struct timeval* nowP, struct stat* sbP );
+extern void mmc_unmap( void* addr, struct stat* sbP, struct timeval* nowP );
 
 /* Clean up the mmc package, freeing any unused storage.
 ** This should be called periodically, say every five minutes.
@@ -48,7 +49,7 @@ extern void mmc_cleanup( struct timeval* nowP );
 /* Free all storage, usually in preparation for exitting. */
 extern void mmc_destroy( void );
 
-/* Return usage stats on the mmc package. */
-extern void mmc_stats( int* activeP, int* freeP );
+/* Generate debugging statistics syslog message. */
+extern void mmc_logstats( long secs );
 
 #endif /* _MMC_H_ */
