@@ -1,6 +1,6 @@
 /* thttpd.c - tiny/turbo/throttling HTTP server
 **
-** Copyright  1995,1998,1999,2000 by Jef Poskanzer <jef@acme.com>.
+** Copyright  1995,1998,1999,2000,2001 by Jef Poskanzer <jef@acme.com>.
 ** All rights reserved.
 **
 ** Redistribution and use in source and binary forms, with or without
@@ -64,7 +64,7 @@ static char* argv0;
 static int debug;
 static int port;
 static char* dir;
-static int do_chroot, no_symlink, do_vhost;
+static int do_chroot, no_log, no_symlink, do_vhost, do_global_passwd;
 static char* cgi_pattern;
 static char* url_pattern;
 static int no_empty_referers;
@@ -81,7 +81,6 @@ typedef struct {
     char* pattern;
     long limit;
     long rate;
-    time_t last_avg;
     off_t bytes_since_avg;
     int num_sending;
     } throttletab;
@@ -109,6 +108,7 @@ typedef struct {
     } connecttab;
 static connecttab* connects;
 static int numconnects, maxconnects;
+static int httpd_conn_count;
 
 /* The connection states. */
 #define CNST_FREE 0
@@ -120,13 +120,11 @@ static int numconnects, maxconnects;
 
 static httpd_server* hs = (httpd_server*) 0;
 int terminate = 0;
-#ifdef STATS_TIME
-time_t stats_time;
+time_t start_time, stats_time;
 long stats_connections, stats_bytes;
 int stats_simultaneous;
-#endif /* STATS_TIME */
 
-static int fdwatch_recompute;
+static int got_usr1;
 
 
 /* Forwards. */
@@ -145,6 +143,7 @@ static void handle_send( connecttab* c, struct timeval* tvP );
 static void handle_linger( connecttab* c, struct timeval* tvP );
 static int check_throttles( connecttab* c );
 static void clear_throttles( connecttab* c, struct timeval* tvP );
+static void update_throttles( ClientData client_data, struct timeval* nowP );
 static void clear_connection( connecttab* c, struct timeval* tvP );
 static void really_clear_connection( connecttab* c, struct timeval* tvP );
 static void idle_read_connection( ClientData client_data, struct timeval* nowP );
@@ -155,6 +154,8 @@ static void occasional( ClientData client_data, struct timeval* nowP );
 #ifdef STATS_TIME
 static void show_stats( ClientData client_data, struct timeval* nowP );
 #endif /* STATS_TIME */
+static void logstats( struct timeval* nowP );
+static void thttpd_logstats( long secs );
 
 
 static void
@@ -172,6 +173,9 @@ handle_hup( int sig )
     {
     FILE* logfp;
 
+    if ( no_log )
+	return;
+
     /* Re-open the log file. */
     if ( logfile != (char*) 0 )
 	{
@@ -190,13 +194,14 @@ handle_hup( int sig )
 static void
 handle_usr1( int sig )
     {
-    terminate = 1;
-    if ( hs != (httpd_server*) 0 )
-	{
-	httpd_server* ths = hs;
-	hs = (httpd_server*) 0;
-	httpd_terminate( ths );
-	}
+    got_usr1 = 1;
+    }
+
+
+static void
+handle_usr2( int sig )
+    {
+    logstats( (struct timeval*) 0 );
     }
 
 
@@ -210,7 +215,7 @@ main( int argc, char** argv )
     char cwd[MAXPATHLEN];
     FILE* logfp;
     int num_ready;
-    int cnum;
+    int cnum, ridx;
     connecttab* c;
     httpd_conn* hc;
     httpd_sockaddr sa4;
@@ -234,7 +239,7 @@ main( int argc, char** argv )
     if ( port <= 0 )
 	{
 	syslog( LOG_CRIT, "illegal port number" );
-	(void) fprintf( stderr, "illegal port number\n" );
+	(void) fprintf( stderr, "%s: illegal port number\n", argv0 );
 	exit( 1 );
 	}
 
@@ -246,27 +251,36 @@ main( int argc, char** argv )
     if ( ! ( gotv4 || gotv6 ) )
 	{
 	syslog( LOG_ERR, "can't find any valid address" );
-	(void) fprintf( stderr, "can't find any valid address\n" );
+	(void) fprintf( stderr, "%s: can't find any valid address\n", argv0 );
 	exit( 1 );
 	}
 
     /* Throttle file. */
     numthrottles = 0;
     maxthrottles = 0;
+    throttles = (throttletab*) 0;
     if ( throttlefile != (char*) 0 )
 	read_throttlefile( throttlefile );
 
     /* Log file. */
     if ( logfile != (char*) 0 )
 	{
-	logfp = fopen( logfile, "a" );
-	if ( logfp == (FILE*) 0 )
+	if ( strcmp( logfile, "/dev/null" ) == 0 )
 	    {
-	    syslog( LOG_CRIT, "%.80s - %m", logfile );
-	    perror( logfile );
-	    exit( 1 );
+	    no_log = 1;
+	    logfp = (FILE*) 0;
+	    }
+	else
+	    {
+	    logfp = fopen( logfile, "a" );
+	    if ( logfp == (FILE*) 0 )
+		{
+		syslog( LOG_CRIT, "%.80s - %m", logfile );
+		perror( logfile );
+		exit( 1 );
+		}
+	    (void) fcntl( fileno( logfp ), F_SETFD, 1 );
 	    }
-	(void) fcntl( fileno( logfp ), F_SETFD, 1 );
 	}
     else
 	logfp = (FILE*) 0;
@@ -276,7 +290,7 @@ main( int argc, char** argv )
     if ( pwd == (struct passwd*) 0 )
 	{
 	syslog( LOG_CRIT, "unknown user - '%.80s'", user );
-	(void) fprintf( stderr, "unknown user - '%s'\n", user );
+	(void) fprintf( stderr, "%s: unknown user - '%s'\n", argv0, user );
 	exit( 1 );
 	}
     uid = pwd->pw_uid;
@@ -392,7 +406,9 @@ main( int argc, char** argv )
     (void) signal( SIGINT, handle_term );
     (void) signal( SIGPIPE, SIG_IGN );          /* get EPIPE instead */
     (void) signal( SIGHUP, handle_hup );
+    got_usr1 = 0;
     (void) signal( SIGUSR1, handle_usr1 );
+    (void) signal( SIGUSR2, handle_usr2 );
 
     /* Initialize the timer package. */
     tmr_init();
@@ -403,24 +419,37 @@ main( int argc, char** argv )
     hs = httpd_initialize(
 	hostname,
 	gotv4 ? &sa4 : (httpd_sockaddr*) 0, gotv6 ? &sa6 : (httpd_sockaddr*) 0,
-	port, cgi_pattern, charset, cwd, logfp, no_symlink, do_vhost,
-	url_pattern, local_pattern, no_empty_referers );
+	port, cgi_pattern, charset, cwd, no_log, logfp, no_symlink, do_vhost,
+	do_global_passwd, url_pattern, local_pattern, no_empty_referers );
     if ( hs == (httpd_server*) 0 )
 	exit( 1 );
 
     /* Set up the occasional timer. */
-    (void) tmr_create(
-	(struct timeval*) 0, occasional, JunkClientData,
-	OCCASIONAL_TIME * 1000L, 1 );
+    if ( tmr_create( (struct timeval*) 0, occasional, JunkClientData, OCCASIONAL_TIME * 1000L, 1 ) == (Timer*) 0 )
+	{
+	syslog( LOG_CRIT, "tmr_create(occasional) failed" );
+	exit( 1 );
+	}
+    if ( numthrottles > 0 )
+	{
+	/* Set up the throttles timer. */
+	if ( tmr_create( (struct timeval*) 0, update_throttles, JunkClientData, THROTTLE_TIME * 1000L, 1 ) == (Timer*) 0 )
+	    {
+	    syslog( LOG_CRIT, "tmr_create(update_throttles) failed" );
+	    exit( 1 );
+	    }
+	}
 #ifdef STATS_TIME
     /* Set up the stats timer. */
-    (void) tmr_create(
-	(struct timeval*) 0, show_stats, JunkClientData, STATS_TIME * 1000L,
-	1 );
-    stats_time = time( (time_t*) 0 );
+    if ( tmr_create( (struct timeval*) 0, show_stats, JunkClientData, STATS_TIME * 1000L, 1 ) == (Timer*) 0 )
+	{
+	syslog( LOG_CRIT, "tmr_create(show_stats) failed" );
+	exit( 1 );
+	}
+#endif /* STATS_TIME */
+    start_time = stats_time = time( (time_t*) 0 );
     stats_connections = stats_bytes = 0L;
     stats_simultaneous = 0;
-#endif /* STATS_TIME */
 
     /* If we're root, try to become someone else. */
     if ( getuid() == 0 )
@@ -468,7 +497,7 @@ main( int argc, char** argv )
     connects = NEW( connecttab, maxconnects );
     if ( connects == (connecttab*) 0 )
 	{
-	syslog( LOG_CRIT, "out of memory" );
+	syslog( LOG_CRIT, "out of memory allocating a connecttab" );
 	exit( 1 );
 	}
     for ( cnum = 0; cnum < maxconnects; ++cnum )
@@ -477,37 +506,20 @@ main( int argc, char** argv )
 	connects[cnum].hc = (httpd_conn*) 0;
 	}
     numconnects = 0;
-    fdwatch_recompute = 1;
+    httpd_conn_count = 0;
+
+    if ( hs != (httpd_server*) 0 )
+	{
+	if ( hs->listen4_fd != -1 )
+	    fdwatch_add_fd( hs->listen4_fd, (void*) 0, FDW_READ );
+	if ( hs->listen6_fd != -1 )
+	    fdwatch_add_fd( hs->listen6_fd, (void*) 0, FDW_READ );
+	}
 
     /* Main loop. */
     (void) gettimeofday( &tv, (struct timezone*) 0 );
     while ( ( ! terminate ) || numconnects > 0 )
 	{
-	if ( fdwatch_recompute )
-	    {
-	    fdwatch_clear();
-	    if ( numconnects > 0 )
-		for ( cnum = 0; cnum < maxconnects; ++cnum )
-		    switch ( connects[cnum].conn_state )
-			{
-			case CNST_READING:
-			case CNST_LINGERING:
-			fdwatch_add_fd( connects[cnum].hc->conn_fd, FDW_READ );
-			break;
-			case CNST_SENDING:
-			fdwatch_add_fd( connects[cnum].hc->conn_fd, FDW_WRITE );
-			break;
-			}
-	    if ( hs != (httpd_server*) 0 )
-		{
-		if ( hs->listen4_fd != -1 )
-		    fdwatch_add_fd( hs->listen4_fd, FDW_READ );
-		if ( hs->listen6_fd != -1 )
-		    fdwatch_add_fd( hs->listen6_fd, FDW_READ );
-		}
-	    fdwatch_recompute = 0;
-	    }
-
 	/* Do the fd watch. */
 	num_ready = fdwatch( tmr_mstimeout( &tv ) );
 	if ( num_ready < 0 )
@@ -526,60 +538,55 @@ main( int argc, char** argv )
 	    }
 
 	/* Is it a new connection? */
-	if ( hs != (httpd_server*) 0 && hs->listen4_fd != -1 && 
-	     fdwatch_check_fd( hs->listen4_fd, FDW_READ ) )
-	    {
-	    --num_ready;
-	    if ( handle_newconnect( &tv, hs->listen4_fd ) )
-		/* Go around the loop and do another fdwatch, rather than
-		** dropping through and processing existing connections.
-		** New connections always get priority.
-		*/
-		continue;
-	    /* Only if handle_newconnect() fails in a particular way do
-	    ** we fall through.
-	    */
-	    }
 	if ( hs != (httpd_server*) 0 && hs->listen6_fd != -1 && 
-	     fdwatch_check_fd( hs->listen6_fd, FDW_READ ) )
+	     fdwatch_check_fd( hs->listen6_fd ) )
 	    {
-	    --num_ready;
 	    if ( handle_newconnect( &tv, hs->listen6_fd ) )
 		/* Go around the loop and do another fdwatch, rather than
 		** dropping through and processing existing connections.
 		** New connections always get priority.
 		*/
 		continue;
-	    /* Only if handle_newconnect() fails in a particular way do
-	    ** we fall through.
-	    */
+	    }
+	if ( hs != (httpd_server*) 0 && hs->listen4_fd != -1 && 
+	     fdwatch_check_fd( hs->listen4_fd ) )
+	    {
+	    if ( handle_newconnect( &tv, hs->listen4_fd ) )
+		/* Go around the loop and do another fdwatch, rather than
+		** dropping through and processing existing connections.
+		** New connections always get priority.
+		*/
+		continue;
 	    }
 
 	/* Find the connections that need servicing. */
-	for ( cnum = 0; num_ready > 0 && cnum < maxconnects; ++cnum )
+	for ( ridx = 0; ridx < num_ready; ++ridx )
 	    {
-	    c = &connects[cnum];
+	    c = (connecttab*) fdwatch_get_client_data( ridx );
+	    if ( c == (connecttab*) 0 )
+		continue;
 	    hc = c->hc;
 	    if ( c->conn_state == CNST_READING &&
-		 fdwatch_check_fd( hc->conn_fd, FDW_READ ) )
-		{
-		--num_ready;
+		 fdwatch_check_fd( hc->conn_fd ) )
 		handle_read( c, &tv );
-		}
 	    else if ( c->conn_state == CNST_SENDING &&
-		 fdwatch_check_fd( hc->conn_fd, FDW_WRITE ) )
-		{
-		--num_ready;
+		 fdwatch_check_fd( hc->conn_fd ) )
 		handle_send( c, &tv );
-		}
 	    else if ( c->conn_state == CNST_LINGERING &&
-		 fdwatch_check_fd( hc->conn_fd, FDW_READ ) )
-		{
-		--num_ready;
+		 fdwatch_check_fd( hc->conn_fd ) )
 		handle_linger( c, &tv );
-		}
 	    }
 	tmr_run( &tv );
+
+	if ( got_usr1 && ! terminate )
+	    {
+	    terminate = 1;
+	    if ( hs != (httpd_server*) 0 )
+		{
+		httpd_terminate( hs );
+		hs = (httpd_server*) 0;
+		}
+	    }
 	}
 
     /* The main loop terminated. */
@@ -603,12 +610,18 @@ parse_args( int argc, char** argv )
 #else /* ALWAYS_CHROOT */
     do_chroot = 0;
 #endif /* ALWAYS_CHROOT */
+    no_log = 0;
     no_symlink = do_chroot;
 #ifdef ALWAYS_VHOST
     do_vhost = 1;
 #else /* ALWAYS_VHOST */
     do_vhost = 0;
 #endif /* ALWAYS_VHOST */
+#ifdef ALWAYS_GLOBAL_PASSWD
+    do_global_passwd = 1;
+#else /* ALWAYS_GLOBAL_PASSWD */
+    do_global_passwd = 0;
+#endif /* ALWAYS_GLOBAL_PASSWD */
 #ifdef CGI_PATTERN
     cgi_pattern = CGI_PATTERN;
 #else /* CGI_PATTERN */
@@ -626,9 +639,7 @@ parse_args( int argc, char** argv )
     argn = 1;
     while ( argn < argc && argv[argn][0] == '-' )
 	{
-	if ( strcmp( argv[argn], "-D" ) == 0 )
-	    debug = 1;
-	else if ( strcmp( argv[argn], "-C" ) == 0 && argn + 1 < argc )
+	if ( strcmp( argv[argn], "-C" ) == 0 && argn + 1 < argc )
 	    {
 	    ++argn;
 	    read_config( argv[argn] );
@@ -686,6 +697,10 @@ parse_args( int argc, char** argv )
 	    do_vhost = 1;
 	else if ( strcmp( argv[argn], "-nov" ) == 0 )
 	    do_vhost = 0;
+	else if ( strcmp( argv[argn], "-g" ) == 0 )
+	    do_global_passwd = 1;
+	else if ( strcmp( argv[argn], "-nog" ) == 0 )
+	    do_global_passwd = 0;
 	else if ( strcmp( argv[argn], "-i" ) == 0 && argn + 1 < argc )
 	    {
 	    ++argn;
@@ -696,6 +711,13 @@ parse_args( int argc, char** argv )
 	    ++argn;
 	    charset = argv[argn];
 	    }
+	else if ( strcmp( argv[argn], "-V" ) == 0 )
+	    {
+	    (void) fprintf( stderr, "%s\n", SERVER_SOFTWARE );
+	    exit( 0 );
+	    }
+	else if ( strcmp( argv[argn], "-D" ) == 0 )
+	    debug = 1;
 	else
 	    usage();
 	++argn;
@@ -709,7 +731,7 @@ static void
 usage( void )
     {
     (void) fprintf( stderr,
-"usage:  %s [-C configfile] [-D] [-p port] [-d dir] [-r|-nor] [-v|-nov] [-u user] [-c cgipat] [-t throttles] [-h host] [-l logfile] [-i pidfile] [-T charset]\n",
+"usage:  %s [-C configfile] [-p port] [-d dir] [-r|-nor] [-v|-nov] [-g|-nog] [-u user] [-c cgipat] [-t throttles] [-h host] [-l logfile] [-i pidfile] [-T charset] [-V] [-D]\n",
 	argv0 );
     exit( 1 );
     }
@@ -851,6 +873,16 @@ read_config( char* filename )
 		no_value_required( name, value );
 		do_vhost = 0;
 		}
+	    else if ( strcasecmp( name, "globalpasswd" ) == 0 )
+		{
+		no_value_required( name, value );
+		do_global_passwd = 1;
+		}
+	    else if ( strcasecmp( name, "noglobalpasswd" ) == 0 )
+		{
+		no_value_required( name, value );
+		do_global_passwd = 0;
+		}
 	    else if ( strcasecmp( name, "pidfile" ) == 0 )
 		{
 		value_required( name, value );
@@ -907,8 +939,8 @@ e_strdup( char* oldstr )
     newstr = strdup( oldstr );
     if ( newstr == (char*) 0 )
 	{
-	syslog( LOG_CRIT, "out of memory" );
-	(void) fprintf( stderr, "%s: out of memory\n", argv0 );
+	syslog( LOG_CRIT, "out of memory copying a string" );
+	(void) fprintf( stderr, "%s: out of memory copying a string\n", argv0 );
 	exit( 1 );
 	}
     return newstr;
@@ -935,8 +967,12 @@ lookup_hostname( httpd_sockaddr* sa4P, size_t sa4_len, int* gotv4P, httpd_sockad
     (void) snprintf( strport, sizeof(strport), "%d", port );
     if ( (gaierr = getaddrinfo( hostname, strport, &hints, &ai )) != 0 )
 	{
-	syslog( LOG_CRIT, "getaddrinfo %.80s - %s", hostname, gai_strerror( gaierr ) );
-	(void) fprintf( stderr, "getaddrinfo %.80s - %s\n", hostname, gai_strerror( gaierr ) );
+	syslog(
+	    LOG_CRIT, "getaddrinfo %.80s - %.80s",
+	    hostname, gai_strerror( gaierr ) );
+	(void) fprintf(
+	    stderr, "%s: getaddrinfo %s - %s\n",
+	    argv0, hostname, gai_strerror( gaierr ) );
 	exit( 1 );
 	}
 
@@ -1018,18 +1054,25 @@ lookup_hostname( httpd_sockaddr* sa4P, size_t sa4_len, int* gotv4P, httpd_sockad
 	    if ( he == (struct hostent*) 0 )
 		{
 #ifdef HAVE_HSTRERROR
-		syslog( LOG_CRIT, "gethostbyname %.80s - %s", hostname, hstrerror( h_errno ) );
-		(void) fprintf( stderr, "gethostbyname %.80s - %s\n", hostname, hstrerror( h_errno ) );
+		syslog(
+		    LOG_CRIT, "gethostbyname %.80s - %.80s",
+		    hostname, hstrerror( h_errno ) );
+		(void) fprintf(
+		    stderr, "%s: gethostbyname %s - %s\n",
+		    argv0, hostname, hstrerror( h_errno ) );
 #else /* HAVE_HSTRERROR */
 		syslog( LOG_CRIT, "gethostbyname %.80s failed", hostname );
-		(void) fprintf( stderr, "gethostbyname %.80s failed\n", hostname );
+		(void) fprintf(
+		    stderr, "%s: gethostbyname %s failed\n", argv0, hostname );
 #endif /* HAVE_HSTRERROR */
 		exit( 1 );
 		}
 	    if ( he->h_addrtype != AF_INET )
 		{
 		syslog( LOG_CRIT, "%.80s - non-IP network address", hostname );
-		(void) fprintf( stderr, "%.80s - non-IP network address\n", hostname );
+		(void) fprintf(
+		    stderr, "%s: %s - non-IP network address\n",
+		    argv0, hostname );
 		exit( 1 );
 		}
 	    (void) memcpy(
@@ -1088,7 +1131,8 @@ read_throttlefile( char* throttlefile )
 	    syslog( LOG_CRIT,
 		"unparsable line in %.80s - %.80s", throttlefile, buf );
 	    (void) fprintf( stderr,
-		"unparsable line in %.80s - %.80s\n", throttlefile, buf );
+		"%s: unparsable line in %.80s - %.80s\n",
+		argv0, throttlefile, buf );
 	    continue;
 	    }
 
@@ -1113,8 +1157,10 @@ read_throttlefile( char* throttlefile )
 		}
 	    if ( throttles == (throttletab*) 0 )
 		{
-		syslog( LOG_CRIT, "out of memory" );
-		(void) fprintf( stderr, "out of memory\n" );
+		syslog( LOG_CRIT, "out of memory allocating a throttletab" );
+		(void) fprintf(
+		    stderr, "%s: out of memory allocating a throttletab\n",
+		    argv0 );
 		exit( 1 );
 		}
 	    }
@@ -1123,13 +1169,14 @@ read_throttlefile( char* throttlefile )
 	throttles[numthrottles].pattern = strdup( pattern );
 	if ( throttles[numthrottles].pattern == (char*) 0 )
 	    {
-	    syslog( LOG_CRIT, "out of memory" );
-	    (void) fprintf( stderr, "out of memory\n" );
+	    syslog( LOG_CRIT, "out of memory copying a throttle pattern" );
+	    (void) fprintf(
+		stderr, "%s: out of memory copying a throttle pattern\n",
+		argv0 );
 	    exit( 1 );
 	    }
 	throttles[numthrottles].limit = limit;
 	throttles[numthrottles].rate = 0;
-	throttles[numthrottles].last_avg = tv.tv_sec;
 	throttles[numthrottles].bytes_since_avg = 0;
 	throttles[numthrottles].num_sending = 0;
 
@@ -1145,16 +1192,19 @@ shut_down( void )
     int cnum;
     struct timeval tv;
 
-#ifdef STATS_TIME
-    show_stats( JunkClientData, (struct timeval*) 0 );
-#endif /* STATS_TIME */
     (void) gettimeofday( &tv, (struct timezone*) 0 );
+    logstats( &tv );
     for ( cnum = 0; cnum < maxconnects; ++cnum )
 	{
 	if ( connects[cnum].conn_state != CNST_FREE )
 	    httpd_close_conn( connects[cnum].hc, &tv );
 	if ( connects[cnum].hc != (httpd_conn*) 0 )
+	    {
 	    httpd_destroy_conn( connects[cnum].hc );
+	    free( (void*) connects[cnum].hc );
+	    --httpd_conn_count;
+	    connects[cnum].hc = (httpd_conn*) 0;
+	    }
 	}
     if ( hs != (httpd_server*) 0 )
 	{
@@ -1165,6 +1215,8 @@ shut_down( void )
     mmc_destroy();
     tmr_destroy();
     free( (void*) connects );
+    if ( throttles != (throttletab*) 0 )
+	free( (void*) throttles );
     }
 
 
@@ -1174,7 +1226,6 @@ handle_newconnect( struct timeval* tvP, int listen_fd )
     int cnum;
     connecttab* c;
     ClientData client_data;
-    int flags;
 
     /* This loops until the accept() fails, trying to start new
     ** connections as fast as possible so we don't overrun the
@@ -1204,10 +1255,11 @@ handle_newconnect( struct timeval* tvP, int listen_fd )
 	    c->hc = NEW( httpd_conn, 1 );
 	    if ( c->hc == (httpd_conn*) 0 )
 		{
-		syslog( LOG_CRIT, "out of memory" );
+		syslog( LOG_CRIT, "out of memory allocating an httpd_conn" );
 		exit( 1 );
 		}
 	    c->hc->initialized = 0;
+	    ++httpd_conn_count;
 	    }
 
 	/* Get the connection. */
@@ -1218,12 +1270,16 @@ handle_newconnect( struct timeval* tvP, int listen_fd )
 	    return 1;
 	    }
 	c->conn_state = CNST_READING;
-	fdwatch_recompute = 1;
 	++numconnects;
 	client_data.p = c;
 	c->idle_read_timer = tmr_create(
 	    tvP, idle_read_connection, client_data, IDLE_READ_TIMELIMIT * 1000L,
 	    0 );
+	if ( c->idle_read_timer == (Timer*) 0 )
+	    {
+	    syslog( LOG_CRIT, "tmr_create(idle_read_connection) failed" );
+	    exit( 1 );
+	    }
 	c->idle_send_timer = (Timer*) 0;
 	c->wakeup_timer = (Timer*) 0;
 	c->linger_timer = (Timer*) 0;
@@ -1231,17 +1287,13 @@ handle_newconnect( struct timeval* tvP, int listen_fd )
 	c->numtnums = 0;
 
 	/* Set the connection file descriptor to no-delay mode. */
-	flags = fcntl( c->hc->conn_fd, F_GETFL, 0 );
-	if ( flags == -1 )
-	    syslog( LOG_ERR, "fcntl F_GETFL - %m" );
-	else if ( fcntl( c->hc->conn_fd, F_SETFL, flags | O_NDELAY ) < 0 )
-	    syslog( LOG_ERR, "fcntl O_NDELAY - %m" );
+	httpd_set_ndelay( c->hc->conn_fd );
+
+	fdwatch_add_fd( c->hc->conn_fd, c, FDW_READ );
 
-#ifdef STATS_TIME
 	++stats_connections;
 	if ( numconnects > stats_simultaneous )
 	    stats_simultaneous = numconnects;
-#endif /* STATS_TIME */
 	}
     }
 
@@ -1270,7 +1322,12 @@ handle_read( connecttab* c, struct timeval* tvP )
     sz = read(
 	hc->conn_fd, &(hc->read_buf[hc->read_idx]),
 	hc->read_size - hc->read_idx );
-    if ( sz <= 0 )
+    /* Ignore EWOULDBLOCK errors.  At first glance you would think that
+    ** connections returned by fdwatch as readable should never give an
+    ** EWOULDBLOCK; however, this apparently can happen if a packet gets
+    ** garbled.
+    */
+    if ( sz == 0 || ( sz < 0 && ( errno != EWOULDBLOCK ) ) )
 	{
 	httpd_send_err( hc, 400, httpd_err400title, "", httpd_err400form, "" );
 	clear_connection( c, tvP );
@@ -1306,7 +1363,7 @@ handle_read( connecttab* c, struct timeval* tvP )
 	}
 
     /* Start the connection going. */
-    if ( httpd_start_request( hc ) < 0 )
+    if ( httpd_start_request( hc, tvP ) < 0 )
 	{
 	/* Something went wrong.  Close down the connection. */
 	clear_connection( c, tvP );
@@ -1320,16 +1377,13 @@ handle_read( connecttab* c, struct timeval* tvP )
 	c->bytes_to_send = hc->end_byte_loc + 1;
 	}
     else
-	c->bytes_to_send = hc->bytes;
-#ifdef STATS_TIME
-    stats_bytes += c->bytes_to_send;
-#endif /* STATS_TIME */
+	c->bytes_to_send = hc->bytes_to_send;
 
     /* Check if it's already handled. */
     if ( hc->file_address == (char*) 0 )
 	{
 	/* No file address means someone else is handling it. */
-	c->bytes_sent = hc->bytes;
+	c->bytes_sent = hc->bytes_sent;
 	clear_connection( c, tvP );
 	return;
 	}
@@ -1342,7 +1396,6 @@ handle_read( connecttab* c, struct timeval* tvP )
 
     /* Cool, we have a valid connection and a file to send to it. */
     c->conn_state = CNST_SENDING;
-    fdwatch_recompute = 1;
     c->started_at = tvP->tv_sec;
     c->wouldblock_delay = 0;
     client_data.p = c;
@@ -1351,6 +1404,14 @@ handle_read( connecttab* c, struct timeval* tvP )
     c->idle_send_timer = tmr_create(
 	tvP, idle_send_connection, client_data, IDLE_SEND_TIMELIMIT * 1000L,
 	0 );
+    if ( c->idle_send_timer == (Timer*) 0 )
+	{
+	syslog( LOG_CRIT, "tmr_create(idle_send_connection) failed" );
+	exit( 1 );
+	}
+
+    fdwatch_del_fd( hc->conn_fd );
+    fdwatch_add_fd( hc->conn_fd, c, FDW_WRITE );
     }
 
 
@@ -1399,10 +1460,15 @@ handle_send( connecttab* c, struct timeval* tvP )
 	*/
 	c->wouldblock_delay += MIN_WOULDBLOCK_DELAY;
 	c->conn_state = CNST_PAUSING;
-	fdwatch_recompute = 1;
+	fdwatch_del_fd( hc->conn_fd );
 	client_data.p = c;
 	c->wakeup_timer = tmr_create(
 	    tvP, wakeup_connection, client_data, c->wouldblock_delay, 0 );
+	if ( c->wakeup_timer == (Timer*) 0 )
+	    {
+	    syslog( LOG_CRIT, "tmr_create(wakeup_connection) failed" );
+	    exit( 1 );
+	    }
 	return;
 	}
     if ( sz < 0 )
@@ -1415,9 +1481,11 @@ handle_send( connecttab* c, struct timeval* tvP )
 	** On some systems, write() occasionally gives an EINVAL.
 	** Dunno why, something to do with the socket going
 	** bad.  Anyway, we don't log those either.
+	**
+	** And ECONNRESET isn't interesting either.
 	*/
-	if ( errno != EPIPE && errno != EINVAL )
-	    syslog( LOG_ERR, "write - %m" );
+	if ( errno != EPIPE && errno != EINVAL && errno != ECONNRESET )
+	    syslog( LOG_ERR, "write - %m sending %.80s", hc->encodedurl );
 	clear_connection( c, tvP );
 	return;
 	}
@@ -1445,6 +1513,7 @@ handle_send( connecttab* c, struct timeval* tvP )
 	}
     /* And update how much of the file we wrote. */
     c->bytes_sent += sz;
+    c->hc->bytes_sent += sz;
 
     /* Are we done? */
     if ( c->bytes_sent >= c->bytes_to_send )
@@ -1462,19 +1531,24 @@ handle_send( connecttab* c, struct timeval* tvP )
     if ( c->limit != THROTTLE_NOLIMIT )
 	{
 	elapsed = tvP->tv_sec - c->started_at;
-	if ( elapsed == 0 || c->bytes_sent / elapsed > c->limit )
+	if ( elapsed == 0 || c->hc->bytes_sent / elapsed > c->limit )
 	    {
 	    c->conn_state = CNST_PAUSING;
-	    fdwatch_recompute = 1;
+	    fdwatch_del_fd( hc->conn_fd );
 	    /* When should we send the next c->limit bytes
 	    ** to get back on schedule?  If less than a second
 	    ** (integer math rounding), use 1/8 second.
 	    */
-	    coast = ( c->bytes_sent + c->limit ) / c->limit - elapsed;
+	    coast = ( c->hc->bytes_sent + c->limit ) / c->limit - elapsed;
 	    client_data.p = c;
 	    c->wakeup_timer = tmr_create(
 		tvP, wakeup_connection, client_data,
 		coast ? ( coast * 1000L ) : 125L, 0 );
+	    if ( c->wakeup_timer == (Timer*) 0 )
+		{
+		syslog( LOG_CRIT, "tmr_create(wakeup_connection) failed" );
+		exit( 1 );
+		}
 	    }
 	}
     }
@@ -1506,16 +1580,16 @@ check_throttles( connecttab* c )
 	  ++tnum )
 	if ( match( throttles[tnum].pattern, c->hc->expnfilename ) )
 	    {
-	    c->tnums[c->numtnums++] = tnum;
 	    /* If we're way over the limit, don't even start. */
-	    if ( throttles[tnum].rate > throttles[tnum].limit * 3 / 2 )
+	    if ( throttles[tnum].rate > throttles[tnum].limit * 2 )
 		return 0;
-	    ++throttles[tnum].num_sending;
-	    if ( throttles[tnum].num_sending <= 0 )
+	    if ( throttles[tnum].num_sending < 0 )
 		{
-		syslog( LOG_ERR, "throttle sending count was negative!" );
-		throttles[tnum].num_sending = 1;
+		syslog( LOG_ERR, "throttle sending count was negative - shouldn't happen!" );
+		throttles[tnum].num_sending = 0;
 		}
+	    c->tnums[c->numtnums++] = tnum;
+	    ++throttles[tnum].num_sending;
 	    c->limit = MIN(
 		c->limit, throttles[tnum].limit / throttles[tnum].num_sending );
 	    }
@@ -1527,21 +1601,34 @@ static void
 clear_throttles( connecttab* c, struct timeval* tvP )
     {
     int i, tnum;
-    time_t elapsed;
 
     for ( i = 0; i < c->numtnums; ++i )
 	{
 	tnum = c->tnums[i];
 	--throttles[tnum].num_sending;
-	throttles[tnum].bytes_since_avg += c->bytes_sent;
-	elapsed = tvP->tv_sec - throttles[tnum].last_avg;
-	if ( elapsed >= THROTTLE_TIME )
+	throttles[tnum].bytes_since_avg += c->hc->bytes_sent;
+	}
+    }
+
+
+static void
+update_throttles( ClientData client_data, struct timeval* nowP )
+    {
+    int tnum;
+
+    for ( tnum = 0; tnum < numthrottles; ++tnum )
+	{
+	throttles[tnum].rate =
+	    ( 3 * throttles[tnum].rate +
+	      throttles[tnum].bytes_since_avg / THROTTLE_TIME ) / 4;
+	throttles[tnum].bytes_since_avg = 0;
+	/* Log a warning message if necessary. */
+	if ( throttles[tnum].rate > throttles[tnum].limit )
 	    {
-	    throttles[tnum].rate =
-		( throttles[tnum].rate +
-		  throttles[tnum].bytes_since_avg / elapsed ) / 2;
-	    throttles[tnum].bytes_since_avg = 0;
-	    throttles[tnum].last_avg = tvP->tv_sec;
+	    if ( throttles[tnum].rate > throttles[tnum].limit * 2 )
+		syslog( LOG_NOTICE, "throttle #%d '%.80s' rate %ld GREATLY exceeding limit %ld", tnum, throttles[tnum].pattern, throttles[tnum].rate, throttles[tnum].limit );
+	    else
+		syslog( LOG_NOTICE, "throttle #%d '%.80s' rate %ld exceeding limit %ld", tnum, throttles[tnum].pattern, throttles[tnum].rate, throttles[tnum].limit );
 	    }
 	}
     }
@@ -1585,10 +1672,18 @@ clear_connection( connecttab* c, struct timeval* tvP )
     if ( c->hc->should_linger )
 	{
 	c->conn_state = CNST_LINGERING;
-	fdwatch_recompute = 1;
+	fdwatch_del_fd( c->hc->conn_fd );
+	fdwatch_add_fd( c->hc->conn_fd, c, FDW_READ );
+	/* Make sure we are still in no-delay mode. */
+	httpd_set_ndelay( c->hc->conn_fd );
 	client_data.p = c;
 	c->linger_timer = tmr_create(
 	    tvP, linger_clear_connection, client_data, LINGER_TIME * 1000L, 0 );
+	if ( c->linger_timer == (Timer*) 0 )
+	    {
+	    syslog( LOG_CRIT, "tmr_create(linger_clear_connection) failed" );
+	    exit( 1 );
+	    }
 	}
     else
 	really_clear_connection( c, tvP );
@@ -1598,6 +1693,8 @@ clear_connection( connecttab* c, struct timeval* tvP )
 static void
 really_clear_connection( connecttab* c, struct timeval* tvP )
     {
+    stats_bytes += c->bytes_to_send;
+    fdwatch_del_fd( c->hc->conn_fd );
     httpd_close_conn( c->hc, tvP );
     clear_throttles( c, tvP );
     if ( c->linger_timer != (Timer*) 0 )
@@ -1606,7 +1703,6 @@ really_clear_connection( connecttab* c, struct timeval* tvP )
 	c->linger_timer = 0;
 	}
     c->conn_state = CNST_FREE;
-    fdwatch_recompute = 1;
     --numconnects;
     }
 
@@ -1656,7 +1752,7 @@ wakeup_connection( ClientData client_data, struct timeval* nowP )
     if ( c->conn_state == CNST_PAUSING )
 	{
 	c->conn_state = CNST_SENDING;
-	fdwatch_recompute = 1;
+	fdwatch_add_fd( c->hc->conn_fd, c, FDW_WRITE );
 	}
     }
 
@@ -1683,23 +1779,50 @@ occasional( ClientData client_data, struct timeval* nowP )
 static void
 show_stats( ClientData client_data, struct timeval* nowP )
     {
-    time_t now;
-    int am, fm, at, ft;
-    long ns, np;
-    long secs;
+    logstats( nowP );
+    }
+#endif /* STATS_TIME */
 
-    now = time( (time_t*) 0 );
-    mmc_stats( &am, &fm );
-    tmr_stats( &at, &ft );
-    fdwatch_stats( &ns, &np );
-    secs = now - stats_time;
-    syslog( LOG_NOTICE,
-	"%ld seconds, %ld connections (%g/sec), %d max simultaneous, %ld bytes (%g/sec), %d maps (%d free), %d timers (%d free), %ld selects, %ld polls",
-	secs, stats_connections, (float) stats_connections / secs,
-	stats_simultaneous, stats_bytes, (float) stats_bytes / secs, am, fm,
-	at, ft, ns, np );
+
+/* Generate debugging statistics syslog messages for all packages. */
+static void
+logstats( struct timeval* nowP )
+    {
+    struct timeval tv;
+    time_t now;
+    long up_secs, stats_secs;
+
+    if ( nowP == (struct timeval*) 0 )
+	{
+	(void) gettimeofday( &tv, (struct timezone*) 0 );
+	nowP = &tv;
+	}
+    now = nowP->tv_sec;
+    up_secs = now - start_time;
+    stats_secs = now - stats_time;
+    if ( stats_secs == 0 )
+	stats_secs = 1;	/* fudge */
     stats_time = now;
+    syslog( LOG_NOTICE,
+	"up %ld seconds, stats for %ld seconds:", up_secs, stats_secs );
+
+    thttpd_logstats( stats_secs );
+    httpd_logstats( stats_secs );
+    mmc_logstats( stats_secs );
+    fdwatch_logstats( stats_secs );
+    tmr_logstats( stats_secs );
+    }
+
+
+/* Generate debugging statistics syslog message. */
+static void
+thttpd_logstats( long secs )
+    {
+    syslog( LOG_NOTICE,
+	"  thttpd - %ld connections (%g/sec), %d max simultaneous, %ld bytes (%g/sec), %d httpd_conns allocated",
+	stats_connections, (float) stats_connections / secs,
+	stats_simultaneous, stats_bytes, (float) stats_bytes / secs,
+	httpd_conn_count );
     stats_connections = stats_bytes = 0L;
     stats_simultaneous = 0;
     }
-#endif /* STATS_TIME */
